@@ -6,17 +6,19 @@ import threading
 import typing
 from contextlib import contextmanager
 from importlib import import_module
+from ipaddress import ip_address, IPv6Address
 from ssl import SSLContext
-from typing import TypeVar, Callable, Union, Optional, Awaitable, Coroutine, Any, Dict
+from typing import TypeVar, Callable, Union, Optional, Awaitable, Coroutine, Any, Dict, List, Tuple
 
 import sniffio
 
 from .abc import (  # noqa: F401
     IPAddressType, CancelScope, UDPSocket, Lock, Condition, Event, Semaphore, Queue, TaskGroup,
-    Stream, SocketStreamServer, SocketStream, AsyncFile)
+    Stream, SocketStreamServer, SocketStream, AsyncFile, CapacityLimiter)
 from . import _networking
 
 BACKENDS = 'asyncio', 'curio', 'trio'
+IPPROTO_IPV6 = getattr(socket, 'IPPROTO_IPV6', 41)  # https://bugs.python.org/issue29515
 
 T_Retval = TypeVar('T_Retval', covariant=True)
 T_Agen = TypeVar('T_Agen')
@@ -213,16 +215,24 @@ def create_task_group() -> TaskGroup:
 # Threads
 #
 
-def run_in_thread(func: Callable[..., T_Retval], *args) -> Awaitable[T_Retval]:
+def run_in_thread(func: Callable[..., T_Retval], *args, cancellable: bool = False,
+                  limiter: Optional[CapacityLimiter] = None) -> Awaitable[T_Retval]:
     """
     Start a thread that calls the given function with the given arguments.
 
+    If the ``cancellable`` option is enabled and the task waiting for its completion is cancelled,
+    the thread will still run its course but its return value (or any raised exception) will be
+    ignored.
+
     :param func: a callable
     :param args: positional arguments for the callable
+    :param cancellable: ``True`` to allow cancellation of the operation
+    :param limiter: capacity limiter to use to limit the total amount of threads running
+        (if omitted, the default limiter is used)
     :return: an awaitable that yields the return value of the function.
 
     """
-    return _get_asynclib().run_in_thread(func, *args)
+    return _get_asynclib().run_in_thread(func, *args, cancellable=cancellable, limiter=limiter)
 
 
 def run_async_from_thread(func: Callable[..., Coroutine[Any, Any, T_Retval]], *args) -> T_Retval:
@@ -242,6 +252,16 @@ def run_async_from_thread(func: Callable[..., Coroutine[Any, Any, T_Retval]], *a
     return asynclib.run_async_from_thread(func, *args)
 
 
+def current_default_thread_limiter() -> CapacityLimiter:
+    """
+    Return the capacity limiter that is used by default to limit the number of concurrent threads.
+
+    :return: a capacity limiter object
+
+    """
+    return _get_asynclib().current_default_thread_limiter()
+
+
 #
 # Async file I/O
 #
@@ -272,11 +292,18 @@ def aopen(file: Union[str, 'os.PathLike', int], mode: str = 'r', buffering: int
 async def connect_tcp(
     address: IPAddressType, port: int, *, ssl_context: Optional[SSLContext] = None,
     autostart_tls: bool = False, bind_host: Optional[IPAddressType] = None,
-    bind_port: Optional[int] = None, tls_standard_compatible: bool = True
+    bind_port: Optional[int] = None, tls_standard_compatible: bool = True,
+    happy_eyeballs_delay: float = 0.25
 ) -> SocketStream:
     """
     Connect to a host using the TCP protocol.
 
+    This function implements the stateless version of the Happy Eyeballs algorithm (RFC 6555).
+    If ``address`` is a host name that resolves to multiple IP addresses, each one is tried until
+    one connection attempt succeeds. If the first attempt does not connected within 250
+    milliseconds, a second attempt is started using the next address in the list, and so on.
+    For IPv6 enabled systems, IPv6 addresses are tried first.
+
     :param address: the IP address or host name to connect to
     :param port: port on the target host to connect to
     :param ssl_context: default SSL context to use for TLS handshakes
@@ -288,29 +315,96 @@ async def connect_tcp(
         :exc:`~ssl.SSLEOFError` may be raised during reads from the stream.
         Some protocols, such as HTTP, require this option to be ``False``.
         See :meth:`~ssl.SSLContext.wrap_socket` for details.
+    :param happy_eyeballs_delay: delay (in seconds) before starting the next connection attempt
     :return: a socket stream object
+    :raises OSError: if the connection attempt fails
 
     """
+    # Placed here due to https://github.com/python/mypy/issues/7057
+    stream = None  # type: Optional[SocketStream]
+
+    async def try_connect(af: int, addr: str, event: Event):
+        nonlocal stream
+        try:
+            raw_socket = socket.socket(af, socket.SOCK_STREAM)
+        except OSError as exc:
+            oserrors.append(exc)
+            await event.set()
+            return
+
+        sock = asynclib.Socket(raw_socket)
+        try:
+            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
+            if interface is not None and bind_port is not None:
+                await sock.bind((interface, bind_port))
+
+            await sock.connect((addr, port))
+        except OSError as exc:
+            oserrors.append(exc)
+            await sock.close()
+            return
+        except BaseException:
+            await sock.close()
+            raise
+        else:
+            assert stream is None
+            stream = _networking.SocketStream(sock, ssl_context, target_host,
+                                              tls_standard_compatible)
+        finally:
+            await event.set()
+
+    asynclib = _get_asynclib()
+    interface, family = None, 0  # type: Optional[str], int
     if bind_host:
-        bind_host = str(bind_host)
+        interface, family, _v6only = await _networking.get_bind_address(bind_host)
 
-    raw_socket = socket.socket()
-    sock = _get_asynclib().Socket(raw_socket)
+    target_host = str(address)
     try:
-        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
-        if bind_host is not None and bind_port is not None:
-            await sock.bind((bind_host, bind_port))
+        addr_obj = ip_address(address)
+    except ValueError:
+        # getaddrinfo() will raise an exception if name resolution fails
+        resolved = await run_in_thread(socket.getaddrinfo, target_host, port, family,
+                                       socket.SOCK_STREAM, cancellable=True)
 
-        await sock.connect((address, port))
-        stream = _networking.SocketStream(sock, ssl_context, str(address), tls_standard_compatible)
+        # Organize the list so that the first address is an IPv6 address (if available) and the
+        # second one is an IPv4 addresses. The rest can be in whatever order.
+        v6_found = v4_found = False
+        target_addrs = []  # type: List[Tuple[socket.AddressFamily, str]]
+        for af, *rest, sa in resolved:
+            if af == socket.AF_INET6 and not v6_found:
+                v6_found = True
+                target_addrs.insert(0, (af, sa[0]))
+            elif af == socket.AF_INET and not v4_found and v6_found:
+                v4_found = True
+                target_addrs.insert(1, (af, sa[0]))
+            else:
+                target_addrs.append((af, sa[0]))
+    else:
+        if isinstance(addr_obj, IPv6Address):
+            target_addrs = [(socket.AF_INET6, addr_obj.compressed)]
+        else:
+            target_addrs = [(socket.AF_INET, addr_obj.compressed)]
 
-        if autostart_tls:
-            await stream.start_tls()
+    oserrors = []  # type: List[OSError]
+    async with create_task_group() as tg:
+        for i, (af, addr) in enumerate(target_addrs):
+            event = create_event()
+            await tg.spawn(try_connect, af, addr, event)
+            async with move_on_after(happy_eyeballs_delay):
+                await event.wait()
 
-        return stream
-    except BaseException:
-        await sock.close()
-        raise
+            if stream is not None:
+                await tg.cancel_scope.cancel()
+                break
+
+    if stream is None:
+        cause = oserrors[0] if len(oserrors) == 1 else asynclib.ExceptionGroup(oserrors)
+        raise OSError('All connection attempts failed') from cause
+
+    if autostart_tls:
+        await stream.start_tls()
+
+    return stream
 
 
 async def connect_unix(path: Union[str, 'os.PathLike']) -> SocketStream:
@@ -342,7 +436,9 @@ async def create_tcp_server(
     Start a TCP socket server.
 
     :param port: port number to listen on
-    :param interface: interface to listen on (if omitted, listen on any interface)
+    :param interface: IP address of the interface to listen on. If omitted, listen on all IPv4
+        and IPv6 interfaces. To listen on all interfaces on a specific address family, use
+        ``0.0.0.0`` for IPv4 or ``::`` for IPv6.
     :param ssl_context: an SSL context object for TLS negotiation
     :param autostart_tls: automatically do the TLS handshake on new connections if ``ssl_context``
         has been provided
@@ -354,10 +450,13 @@ async def create_tcp_server(
     :return: a server object
 
     """
-    if interface:
-        interface = str(interface)
+    interface, family, v6only = await _networking.get_bind_address(interface)
+    raw_socket = socket.socket(family)
+
+    # Enable/disable dual stack operation as needed
+    if family == socket.AF_INET6:
+        raw_socket.setsockopt(IPPROTO_IPV6, socket.IPV6_V6ONLY, v6only)
 
-    raw_socket = socket.socket()
     sock = _get_asynclib().Socket(raw_socket)
     try:
         if sys.platform == 'win32':
@@ -411,7 +510,7 @@ async def create_udp_socket(
     If ``port`` has been given, the socket will be bound to this port on the local machine,
     making this socket suitable for providing UDP based services.
 
-    :param interface: interface to bind to
+    :param interface: IP address of the interface to bind to
     :param port: port to bind to
     :param target_host: remote host to set as the default target
     :param target_port: port on the remote host to set as the default target
@@ -419,11 +518,19 @@ async def create_udp_socket(
 
     """
     if interface:
-        interface = str(interface)
-    if target_host:
-        target_host = str(target_host)
+        interface, family, _v6only = await _networking.get_bind_address(interface)
+    else:
+        interface, family = None, 0
 
-    raw_socket = socket.socket(type=socket.SOCK_DGRAM)
+    if target_host:
+        res = await run_in_thread(socket.getaddrinfo, target_host, target_port, family)
+        if res:
+            family, type_, proto, _cn, sa = res[0]
+            target_host, target_port = sa[:2]
+        else:
+            raise ValueError('{!r} cannot be resolved to an IP address'.format(target_host))
+
+    raw_socket = socket.socket(family=family, type=socket.SOCK_DGRAM)
     sock = _get_asynclib().Socket(raw_socket)
     try:
         if interface is not None or port is not None:
@@ -493,14 +600,15 @@ def create_lock() -> Lock:
     return _get_asynclib().Lock()
 
 
-def create_condition() -> Condition:
+def create_condition(lock: Lock = None) -> Condition:
     """
     Create an asynchronous condition.
 
+    :param lock: the lock to base the condition object on
     :return: a condition object
 
     """
-    return _get_asynclib().Condition()
+    return _get_asynclib().Condition(lock=lock)
 
 
 def create_event() -> Event:
@@ -528,13 +636,25 @@ def create_queue(capacity: int) -> Queue:
     """
     Create an asynchronous queue.
 
-    :param capacity: maximum number of items the queue will be able to store
+    :param capacity: maximum number of items the queue will be able to store (0 = infinite)
     :return: a queue object
 
     """
     return _get_asynclib().Queue(capacity)
 
 
+def create_capacity_limiter(total_tokens: float) -> CapacityLimiter:
+    """
+    Create a capacity limiter.
+
+    :param total_tokens: the total number of tokens available for borrowing (can be an integer or
+        :data:`math.inf`)
+    :return: a capacity limiter object
+
+    """
+    return _get_asynclib().CapacityLimiter(total_tokens)
+
+
 #
 # Operating system signals
 #
