@@ -1,11 +1,14 @@
+import asyncio
+
 import curio
 import pytest
 import trio
+from async_generator import async_generator, yield_
 
+import anyio
 from anyio import (
     create_task_group, sleep, move_on_after, fail_after, open_cancel_scope, wait_all_tasks_blocked,
     current_effective_deadline, current_time, get_cancelled_exc_class)
-from anyio._backends import _asyncio
 from anyio.exceptions import ExceptionGroup
 
 
@@ -41,20 +44,22 @@ async def test_success():
     assert results == {'a', 'b'}
 
 
-@pytest.mark.parametrize('run_func, as_coro_obj', [
-    (_asyncio.native_run, True),
-    (curio.run, False),
-    (trio.run, False)
+@pytest.mark.parametrize('module, as_coro_obj', [
+    pytest.param(asyncio, True, marks=[
+        pytest.mark.skipif(not hasattr(asyncio, 'run'), reason='asyncio.run() is not available')]
+    ),
+    (curio, False),
+    (trio, False)
 ], ids=['asyncio', 'curio', 'trio'])
-def test_run_natively(run_func, as_coro_obj):
+def test_run_natively(module, as_coro_obj):
     async def testfunc():
         async with create_task_group() as tg:
             await tg.spawn(sleep, 0)
 
     if as_coro_obj:
-        run_func(testfunc())
+        module.run(testfunc())
     else:
-        run_func(testfunc)
+        module.run(testfunc)
 
 
 @pytest.mark.anyio
@@ -466,3 +471,54 @@ async def test_nested_shield():
                     await tg.spawn(killer, scope)
                     async with fail_after(0.2):
                         await sleep(2)
+
+
+def test_task_group_in_generator(anyio_backend):
+    @async_generator
+    async def task_group_generator():
+        async with create_task_group():
+            await yield_()
+
+    gen = task_group_generator()
+    anyio.run(gen.__anext__, backend=anyio_backend)
+    pytest.raises(StopAsyncIteration, anyio.run, gen.__anext__, backend=anyio_backend)
+
+
+@pytest.mark.anyio
+async def test_exception_group_filtering():
+    """Test that CancelledErrors are filtered out of nested exception groups."""
+
+    async def fail(name):
+        try:
+            await anyio.sleep(.1)
+        finally:
+            raise Exception('%s task failed' % name)
+
+    async def fn():
+        async with anyio.create_task_group() as task_group:
+            await task_group.spawn(fail, 'parent')
+            async with anyio.create_task_group() as task_group2:
+                await task_group2.spawn(fail, 'child')
+                await anyio.sleep(1)
+
+    with pytest.raises(anyio.exceptions.ExceptionGroup) as exc:
+        await fn()
+
+    assert len(exc.value.exceptions) == 2
+    assert str(exc.value.exceptions[0]) == 'parent task failed'
+    assert str(exc.value.exceptions[1]) == 'child task failed'
+
+
+@pytest.mark.anyio
+async def test_cancel_propagation_with_inner_spawn():
+    async def g():
+        async with anyio.create_task_group() as g:
+            await g.spawn(anyio.sleep, 10)
+            await anyio.sleep(5)
+
+        assert False
+
+    async with anyio.create_task_group() as group:
+        await group.spawn(g)
+        await anyio.sleep(0.1)
+        await group.cancel_scope.cancel()
